<?php
/*
	© 2011-2012 Alejandro Exojo Piqueras (aep@exojo.org)

	This file is part of WP Digital Signature.

	WP Digital Signature is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	WP Digital Signature is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with WP Digital Signature.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
	Plugin Name: Digital signature
	Plugin URI: http://example.org
	Description: Provides the capability of verifying the X.509 signature of comments and posts.
	Author: Alejandro Exojo
	Version: 0.1
	Author URI: http://barnacity.net
	License: GPL version 3
*/

// ┏━┓╻╺┳╸┏━╸╻ ╻╻╺┳┓┏━╸   ┏━╸┏━┓┏┓╻┏━╸╻┏━╸╻ ╻┏━┓┏━┓╺┳╸╻┏━┓┏┓╻
// ┗━┓┃ ┃ ┣╸ ┃╻┃┃ ┃┃┣╸    ┃  ┃ ┃┃┗┫┣╸ ┃┃╺┓┃ ┃┣┳┛┣━┫ ┃ ┃┃ ┃┃┗┫
// ┗━┛╹ ╹ ┗━╸┗┻┛╹╺┻┛┗━╸   ┗━╸┗━┛╹ ╹╹  ╹┗━┛┗━┛╹┗╸╹ ╹ ╹ ╹┗━┛╹ ╹

//
// The next blocks create: first, a new page under the Settings group in the
// admin panel; second, a new settings section in that page, and in that
// section, the field that corresponds to the site additional CA.
//

add_action('admin_menu', 'digital_signature_init_options_page');

function digital_signature_init_options_page()
{
	add_options_page(
		_('Digital Signature options'), // Page title.
		_('Digital Signature'), // Text displayed as submenu.
		// WordPress capability that the user has to have to access the menu.
		'manage_options',
		'digital_signature', // Unique "menu slug" to refer the submenu.
		'digital_signature_options_page'); // Callback.
}

function digital_signature_options_page()
{
	echo("<div class=\"wrap\">\n");
	screen_icon();
	echo("<h2>Digital Signature</h2>\n"
		."<form action=\"options.php\" method=\"post\">\n");
	settings_fields('digital_signature_certification_authority');
	do_settings_sections('digital_signature');
	submit_button();
	echo("</form>\n</div>\n");
}

add_action('admin_init', 'digital_signature_admin_init');

function digital_signature_admin_init()
{
	register_setting('digital_signature_certification_authority', // Group name.
		'digital_signature_certification_authority', // Option name.
		'digital_signature_validate_certification_authority'); //Validation func.
	add_settings_section(
		'digital_signature_certification_authority', // HTML id for the section.
		_('Additional Certification Authority'), // Section title (h3 element).
		'digital_signature_options_description', // callback that prints explanation.
		'digital_signature'); // Page "slug"
	add_settings_field(
		// HTML id for the section.
		'digital_signature_ca',
		// Text printed next to the field.
		_('Certification Authority'),
		// Callback function that will echo the form.
		'digital_signature_options_ca_form',
		// Settings page on which the section will be.
		'digital_signature',
		// Section of the settings page in which to show the field (cf.
		// add_settings_section).
		'digital_signature_certification_authority');
}

function digital_signature_options_description()
{
	_e('<p>You can add a custom Certification Authority that will be used in the
		verification of the digital signatures of the content on the site. This
		useful if you want your users to sign the contents, but you don\'t want
		to force them to pay to a root Certification Authority. You can create
		your own, and then sign their certificates, but then is your
		responsibility to verify their identity.</p>');
}

function digital_signature_options_ca_form()
{
	// Output the form itself using builtin WordPress functions.
	$o = get_option('digital_signature_certification_authority');
	$ca = $o['digital_signature_certification_authority'];
	echo('<textarea id="digital_signature_certification_authority" '
		. 'name="digital_signature_certification_authority[digital_signature_certification_authority]" '
		. 'cols="66" rows="20" style="white-space: pre; font-family: monospace;'
		. 'font-size: smaller">'
		. $ca
		. "</textarea>\n");
}

function digital_signature_validate_certification_authority($in)
{
	return $in;
}

// ╻ ╻┏━┓┏━╸┏━┓   ┏━╸┏━┓┏┓╻┏━╸╻┏━╸╻ ╻┏━┓┏━┓╺┳╸╻┏━┓┏┓╻
// ┃ ┃┗━┓┣╸ ┣┳┛   ┃  ┃ ┃┃┗┫┣╸ ┃┃╺┓┃ ┃┣┳┛┣━┫ ┃ ┃┃ ┃┃┗┫
// ┗━┛┗━┛┗━╸╹┗╸   ┗━╸┗━┛╹ ╹╹  ╹┗━┛┗━┛╹┗╸╹ ╹ ╹ ╹┗━┛╹ ╹

//
// Add a textarea in the user profile page for the user's public key.
//
add_action('show_user_profile', 'digital_signature_user_key_field');

function digital_signature_user_key_field($user)
{
	$key = get_user_meta($user->ID, 'public_x509_key');

	echo "<table class=\"form-table\">\n<tbody>\n<tr>\n<th>" .
		 '<label for="public-key">' .
		_('Your ASCII-armored X.509 digital key') .
		"</label>\n</th>" .
		'<td>' .
		// TODO: add this styles in a proper way. The default width is limited
		// to 500px by #profile-page .form-table textarea (by some
		// load-styles.php). Making the font smaller might be enough for now.
		'<textarea id="public-key" name="public-key" cols="64" rows="20" ' .
		'style="white-space: pre; font-family: monospace; font-size: smaller">' .
		// If some key was retrieved from the user profile in the DB, output it.
		( isset($key[0]) ? $key[0] : '' ) .
		"</textarea>\n" .
		"<br>\n" .
		'<span class="description">If you have an X.509 digital certificate
		signed by a valid Certification Authority (or a Certification Authority
		that this website accepts) you can add to your profile the public key
		from that certificate. That public key has to be in <a
		href="http://en.wikipedia.org/wiki/Privacy_Enhanced_Mail">Privacy
		Enhanced Mail</a> format to be saved here. Normally is a file that ends
		in <kbd>.pem</kbd> You will recognize it because it starts with
		<code>-----BEGIN CERTIFICATE-----</code>.</span>' .
		'</td>' .
		'</tr>' .
		'</tbody></table>';
}

//
// When the user profile is saved, check if the digital key should be updated.
//
add_action('personal_options_update', 'digital_signature_user_key_save');

function digital_signature_user_key_save($id)
{
	if ( isset( $_POST['public-key'] ) )
	{
		// TODO: check if the key is valid and return an error if not

		if ( $key = $_POST['public-key'] )
			update_user_meta( $id, 'public_x509_key', $key );
		else
			delete_user_meta( $id, 'public_x509_key');
	}
}


// ┏━╸┏━┓┏┓╻╺┳╸┏━╸┏┓╻╺┳╸   ╻┏┓╻╺┳╸┏━╸┏━┓┏━╸┏━┓┏━╸┏━╸
// ┃  ┃ ┃┃┗┫ ┃ ┣╸ ┃┗┫ ┃    ┃┃┗┫ ┃ ┣╸ ┣┳┛┣╸ ┣━┫┃  ┣╸
// ┗━╸┗━┛╹ ╹ ╹ ┗━╸╹ ╹ ╹    ╹╹ ╹ ╹ ┗━╸╹┗╸╹  ╹ ╹┗━╸┗━╸

//
// Add to the post, page and comment forms a "metabox". That is WordPress
// interface for adding extra fields to the add and edit interface.
//
add_action('add_meta_boxes', 'digital_signature_add_metaboxes'); // WP >3.0

function digital_signature_add_metaboxes()
{
	// Parameters for add_meta_box:
	// - HTML id attribute of the edit screen section.
	// - Title visible to the user (head of the metabox).
	// - Callback that prints the HTML that will be shown to the user.
	// - type (post, page, custom, etc.).
	// - Context (normal, advanced, side).
	// - Oters (optional).
	add_meta_box('digital-signature', _('Digital signature'),
		'digital_signature_post_metabox', 'post', normal);
	add_meta_box('digital-signature', _('Digital signature'),
		'digital_signature_post_metabox', 'page', normal);
	add_meta_box('digital-signature', _('Digital signature'),
		'digital_signature_post_metabox', 'comment', normal);
}

function digital_signature_post_metabox($data)
{
	// This function can be called to add a metabox in different places. This
	// system is used to either create a new post, page, etc., or to edit a
	// post, page _or comment_ (to create a new comment, a different mechanism
	// for adding a field to a form is used, see below). So we have to check if
	// is called when editing a comment, to retrieve the signature properly.

	// TODO: Some sanity checks (nonces, auth, etc.) are used in this example:
	// http://codex.wordpress.org/Function_Reference/add_meta_box

	global $user_ID;

	if ($data->ID) // Is a post, page, etc., but not a comment.
	{
		// Retrieve the public key, if there is any.
		$key = get_user_meta($user_ID, 'public_x509_key');

		if ( ! isset($key[0]) )
		{
			_e('Add your public key from a X.509 certificate to your profile to
				digitally sign this content.');
			return;
		}

		// Retrieve the digital signature, if exists
		$signed    = get_post_meta( $data->ID, '_digital_signature_present');
		$signature = get_post_meta( $data->ID, '_digital_signature_value');

		// Output the form
		echo('<p><input type="checkbox" name="digital-signature-present" ' .
			( $signed == true ? 'checked="checked"' : '' ) . ' />' .
			_('Sign the post contents with your digital certificate.') .
			'<input type="hidden" name="digital-signature-value" ' .
			( $signed == true ? "value=\"$signature[0]\"" : 'value=""') .' />'.
			'<input type="button" name="digital-signature-update" value="' .
			_('Update signature') . '" />' . "</p>\n");
	}
	else if ($data->comment_ID) // Is a comment
	{
		// Retrieve the digital signature, if exists
		$signed = get_comment_meta( $data->comment_ID,
			'_digital_signature_present');
		$signature = get_comment_meta( $data->comment_ID,
			'_digital_signature_value');

		// Output the form
		echo('<p><input type="checkbox" name="digital-signature-present" ' .
			( $signed == true ? 'checked="checked"' : '' ) . ' />' .
			_('Sign the comment with your digital certificate.') .
			'<input type="hidden" name="digital-signature-value" ' .
			( $signed == true ? "value=\"$signature[0]\"" : 'value=""') .' />'.
			'<input type="button" name="digital-signature-update" value="' .
			_('Update signature') . '" />' . "</p>\n");
		// NOTE: post/comment saving are handled at different places
	}
	// TODO: Should this case be checked too?
}

// Hook into save_post to store the signature or delete it.
add_action('save_post', 'digital_signature_post_save');

function digital_signature_post_save($post_id)
{
	if ( isset( $_POST['digital-signature-present'] ) &&
		 isset( $_POST['digital-signature-value'] ) ) {
		update_post_meta($post_id, '_digital_signature_value',
			 $_POST['digital-signature-value']);
		update_post_meta($post_id, '_digital_signature_present', true);
	}
	else {
		delete_post_meta( $post_id, '_digital_signature_value');
		delete_post_meta( $post_id, '_digital_signature_present');
	}
}

// Filter the post/page content to add the status of the signature.
add_filter('the_content', digital_signature_post_output);

function digital_signature_post_output($content)
{
	// if ( is_single() || is_page() )  ??
	// Check if the content is being displayed on a single post or a page.
	// If we’re on some different page, such as the archives or front page, we
	// should simply return the content untouched.
	$id = get_the_ID();
	if ( get_post_meta( $id, '_digital_signature_present') )
		return $content . digital_signature_verify_post($id);
	else
		return $content;
}

function digital_signature_verify_post($post)
{
	$uid = digital_signature_save_helper_files("post", $post);

	$result = digital_signature_invoke_openssl("/tmp/$uid.content.txt",
		"/tmp/$uid.content.pkcs7", "/tmp/$uid.ca.pem");

	digital_signature_clean_helper_files($uid);

	if ($result == 0) {
		$message = '<div class="digital-signature-result digital-signature-ok">'
			.'<p>This message is digitally signed and the signature is verified'
			.' by this website.</p></div>';
	}
	else {
		$message = '<div class="digital-signature-result digital-signature-ko">'
			.'<p>This message is digitally signed, but the signature could not '
			.' be verified by this website. There is either a system problem, '
			.'or the message has been manipulated after the author saved it'
			.'</p></div>';
	}
	return $message;
}

// ┏━╸┏━┓┏┳┓┏┳┓┏━╸┏┓╻╺┳╸┏━┓
// ┃  ┃ ┃┃┃┃┃┃┃┣╸ ┃┗┫ ┃ ┗━┓
// ┗━╸┗━┛╹ ╹╹ ╹┗━╸╹ ╹ ╹ ┗━┛

// Interface: add the extra fields to the form.
add_action('comment_form_defaults', 'digital_signature_comment_field');

function digital_signature_comment_field($defaults)
{
	// NOTE: This applies to the add comment interface; the edit one is added
	// as a meta_box like with posts.

	global $user_ID;

	if ($user_ID == 0) // Anonymous user? Nothing to add then
		return $defaults;

	// Check if the user has digital key
	$key = get_user_meta($user_ID, 'public_x509_key');

	if ( isset($key[0]) )
	{
		$defaults['comment_notes_after']=
			'<p class="comment-form-signature-use"><label>' .
			'<input type="checkbox" name="digital-signature-present" />' .
			_('Sign the comment with your digital certificate.') . '</label>' .
			'<input type="hidden" name="digital-signature-value" value="" /></p>' .
			'<p><input type="button" name="digital-signature-update" value="' .
			_('Update signature') . '" />' . "</p>\n";
	}
	else
	{
		$defaults['comment_notes_after']=
			'<p class="comment-form-signature-use">' .
			_('Add your public key from a X.509 certificate to your profile to
			digitally sign this content.') .
			"</p>\n";
	}

	return $defaults;
}

add_action('comment_post', 'digital_signature_comment_save');

function digital_signature_comment_save($comment_id)
{
	if (isset( $_POST['digital-signature-present'] ) &&
		isset( $_POST['digital-signature-value'] ) ) {
		add_comment_meta( $comment_id, '_digital_signature_present',
			true, true); // Set the value as "true"; make the value unique.
		add_comment_meta( $comment_id, '_digital_signature_value',
			$_POST['digital-signature-value'], true);
	}
}

add_action('comment_edit_redirect', 'digital_signature_comment_edit_save', 10, 2);

function digital_signature_comment_edit_save($redirect, $comment_id)
{
	if ( isset( $_POST['digital-signature-present'] ) &&
		!empty( $_POST['digital-signature-value'] ) ) {
		update_comment_meta( $comment_id, '_digital_signature_present', true);
		update_comment_meta( $comment_id, '_digital_signature_value',
			$_POST['digital-signature-value']);
	}
	else {
		delete_comment_meta( $comment_id, '_digital_signature_present');
		delete_comment_meta( $comment_id, '_digital_signature_value');
	}
	return $redirect;
}

add_filter('comment_text', 'digital_signature_comment_display');

function digital_signature_comment_display($content)
{
	global $comment;

	$s = get_comment_meta( $comment->comment_ID, '_digital_signature_present');
	if ($s[0] == true) { // In the DB is saved as the '1' string.
		$uid = digital_signature_save_helper_files('comment', $comment->comment_ID);
		$result = digital_signature_invoke_openssl("/tmp/$uid.content.txt",
			"/tmp/$uid.content.pkcs7", "/tmp/$uid.ca.pem");
		digital_signature_clean_helper_files($uid);
		if ($result == 0) {
			$content.='<div class="digital-signature-result digital-signature-ok">'
				.'<p>This message is digitally signed and the signature is verified'
				.' by this website.</p></div>';
		}
		else {
			$content.='<div class="digital-signature-result digital-signature-ko">'
				.'<p>This message is digitally signed, but the signature could not '
				.' be verified by this website. There is either a system problem, '
				.'or the message has been manipulated after the author saved it'
				.'</p></div>';
		}
	}
	return $content;
}

// ┏━┓╺┳╸╻ ╻╻  ┏━╸   ┏━┓┏┓╻╺┳┓   ┏━┓┏━╸┏━┓╻┏━┓╺┳╸┏━┓
// ┗━┓ ┃ ┗┳┛┃  ┣╸    ┣━┫┃┗┫ ┃┃   ┗━┓┃  ┣┳┛┃┣━┛ ┃ ┗━┓
// ┗━┛ ╹  ╹ ┗━╸┗━╸   ╹ ╹╹ ╹╺┻┛   ┗━┛┗━╸╹┗╸╹╹   ╹ ┗━┛

add_action('wp_enqueue_scripts', 'digital_signature_add_styles');

function digital_signature_add_styles()
{
	// $url = plugins_url('ds-styles.css', __FILE__);
	$d = dirname(__FILE__);
	$d = explode('/', $d);
	$d = end($d);
	wp_register_style('digital_signature', WP_CONTENT_URL . "/plugins/$d/styles.css");
	wp_enqueue_style( 'digital_signature');
}

add_action('init', 'digital_signature_add_javascript');

function digital_signature_add_javascript()
{
	// Find the name of the directory where this file is, because the JS file
	// will be in the same one.
	$d = dirname(__FILE__);
	$d = explode('/', $d);
	$d = end($d);
	wp_enqueue_script('digitallysign',
		WP_CONTENT_URL . "/plugins/$d/digitallysign.js", array('jquery'));
}


// ┏━┓┏━┓┏━╸┏┓╻┏━┓┏━┓╻
// ┃ ┃┣━┛┣╸ ┃┗┫┗━┓┗━┓┃
// ┗━┛╹  ┗━╸╹ ╹┗━┛┗━┛┗━╸

// Helper functions that only deal with OpenSSL and necessary temporary files.

function digital_signature_invoke_openssl($content, $signature, $ca)
{
	$command =
		"openssl smime -verify "
		."-in $signature "
		."-inform PEM "
		. (isset($ca) ? "-CAfile $ca " : "")
		."-content $content";
	exec($command, $output, $result);
	return $result; // $output?
}

function digital_signature_save_helper_files($type, $id)
{
	if ($type==='post') {
		$signature = get_post_meta($id, '_digital_signature_value');
		$content = get_post($id);
		$content = $content->post_content;
	}
	else if ($type==='comment') {
		$signature = get_comment_meta($id, '_digital_signature_value');
		$content = get_comment($id);
		$content = $content->comment_content;
	}

	// Random number to avoid name clashes in multiple invocations of the page.
	$uid = 'wp_digital_signature' . rand(0, 99999);

	// Content file.
	$fn = "/tmp/$uid.content.txt";
	if (file_exists($fn))
		unlink($fn);
	$f=fopen($fn, "w");
	fwrite($f, str_replace("\r", "", $content));
	fclose($f);

	// PKCS#7 detached signature.
	$fn = "/tmp/$uid.content.pkcs7";
	if (file_exists($fn))
		unlink($fn);
	$f=fopen($fn, "w");
	fwrite($f,
		"-----BEGIN PKCS7-----\r\n" .
		$signature[0]
		. "\r\n-----END PKCS7-----"
	);
	fclose($f);

	// CA file.
	$o = get_option('digital_signature_certification_authority');
	$ca = $o['digital_signature_certification_authority'];
	$fn = "/tmp/$uid.ca.pem";
	if (file_exists($fn))
		unlink($fn);
	$f=fopen($fn, "w");
	fwrite($f,$ca);
	fclose($f);

	return $uid;
}

function digital_signature_clean_helper_files($uid)
{
	unlink("/tmp/$uid.content.txt");
	unlink("/tmp/$uid.content.pkcs7");
	unlink("/tmp/$uid.ca.pem");
}

// vim:cms=//%s:

